"""
Copyright (c) 2025, Auorui.
All rights reserved.

This module is used for reading and displaying images and videos. The image contains
OpenCV and PIL, and the loading method for the video is OpenCV.
"""
from .video_use import VideoCap
from .matplot import imshowplt, StackedpltV1, StackedpltV2, matplotlib_patch
from .readio import read_bgr, read_gray, read_rgb, read_tensor, read_image, imwrite, display,\
    StackedImagesV1, StackedImagesV2, imattributes
from .imtensor import (
    image_to_bchw, bchw_to_image, tensor_to_image, image_to_tensor, imagelist_to_tensor
)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

The following citation imports roughly follow the sorting by year
"""
# 1990s
from .lenet import LeNet

# 2010s
from .alexnet import AlexNet, ZFNet
from .vgg import VGG, vgg11_bn, vgg13_bn, vgg16_bn, vgg19_bn
from .googlenet import GoogLeNet
from .resnet import ResNet, resnet18, resnet34, resnet50, resnet101, resnet152
from .squeezenet import SqueezeNet, squeezenet1_0, squeezenet1_1
from .darknet import darknet19, darknet53
from .mobilenet import MobileNetV1, MobileNetV2, MobileNetV3, MobileNetV3_Large, MobileNetV3_Small
from .densenet import DenseNet, densenet121, densenet161, densenet169, densenet201
from .mnasnet import MNASNet, mnasnet0_5, mnasnet0_75, mnasnet1_0, mnasnet1_3
from .shufflenet import ShuffleNetV1, ShuffleNetV2, shufflenet_v1_g1, shufflenet_v1_g2, \
    shufflenet_v1_g3, shufflenet_v1_g4, shufflenet_v1_g8, shufflenet_v2_x0_5, shufflenet_v2_x1_0, \
    shufflenet_v2_x1_5, shufflenet_v2_x2_0
from .xception import Xception
from .drn import DRN, DRN_A, drn_a_50, drn_c_26, drn_c_42, drn_c_58, drn_d_22, drn_d_24, drn_d_38, \
    drn_d_40, drn_d_54, drn_d_56, drn_d_105, drn_d_107

# 2020s
from .conv2former import conv2former_n, conv2former_t, conv2former_s, conv2former_l, conv2former_b
from .ghostnet import GhostNet, GGhostRegNet, ghostnetv1, ghostnetv2, g_ghost_regnetx_002, \
    g_ghost_regnetx_004, g_ghost_regnetx_006, g_ghost_regnetx_008, g_ghost_regnetx_016, \
    g_ghost_regnetx_032, g_ghost_regnetx_040, g_ghost_regnetx_064, g_ghost_regnetx_080, \
    g_ghost_regnetx_120, g_ghost_regnetx_160, g_ghost_regnetx_320
from .regnet import RegNet, regnetx_002, regnetx_004, regnetx_006, regnetx_008, \
    regnetx_016, regnetx_032, regnetx_040, regnetx_064, regnetx_080, regnetx_120, regnetx_160, regnetx_320
from .fasternet import FasterNet, fasternet_t0, fasternet_t1, fasternet_t2, fasternet_s, \
    fasternet_m, fasternet_l

# vision Transformer
from .vision_transformer import VisionTransformer, vit_b_16, vit_b_32, vit_l_16, vit_l_32, vit_h_14
from .swin_transformer import SwinTransformer, swin_t, swin_s, swin_b, swin_l

se_resnet18 = lambda num_classes, **kwargs: resnet18(num_classes=num_classes, use_se=True, **kwargs)
se_resnet34 = lambda num_classes, **kwargs: resnet34(num_classes=num_classes, use_se=True, **kwargs)
se_resnet50 = lambda num_classes, **kwargs: resnet50(num_classes=num_classes, use_se=True, **kwargs)
se_resnet101 = lambda num_classes, **kwargs: resnet101(num_classes=num_classes, use_se=True, **kwargs)
se_resnet152 = lambda num_classes, **kwargs: resnet152(num_classes=num_classes, use_se=True, **kwargs)






"""
Copyright (c) 2023, Auorui.
All rights reserved.

Image enhancement application
"""
import numpy as np
import random
from PIL import Image
import PIL

class Images:
    if PIL.__version__ >= "10.0.0":
        FLIP_LEFT_RIGHT = Image.Transpose.FLIP_LEFT_RIGHT
        FLIP_TOP_BOTTOM = Image.Transpose.FLIP_TOP_BOTTOM
        PERSPECTIVE = Image.Transform.PERSPECTIVE
        AFFINE = Image.Transform.AFFINE
        NEAREST = Image.Resampling.NEAREST
        ANTIALIAS = Image.Resampling.LANCZOS
        LINEAR = Image.Resampling.BILINEAR
        CUBIC = Image.Resampling.BICUBIC
    else:
        FLIP_LEFT_RIGHT = Image.FLIP_LEFT_RIGHT
        FLIP_TOP_BOTTOM = Image.FLIP_TOP_BOTTOM
        PERSPECTIVE = Image.PERSPECTIVE
        AFFINE = Image.AFFINE
        NEAREST = Image.NEAREST
        ANTIALIAS = Image.ANTIALIAS
        LINEAR = Image.LINEAR
        CUBIC = Image.CUBIC


def random_apply(img, transforms:list, prob):
    """
    以给定的概率随机应用transforms列表
    Args:
        img: Image to be randomly applied a list transformations.
        transforms (list): List of transformations to be applied.
        prob (float): The probability to apply the transformation list.

    Returns:
        Transformed image.
    """
    if prob < random.random():
        return img
    for transform in transforms:
        img = transform(img)
    return img

def random_order(img, transforms:list):
    """
    以随机顺序应用transforms列表.
    Args:
        img: Image to be applied transformations in a random order.
        transforms (list): List of the transformations to be applied.

    Returns:
        Transformed image.
    """
    random.shuffle(transforms)
    for transform in transforms:
        img = transform(img)
    return img

def random_choice(img, transforms:list):
    """
    从transforms列表中随机选择一个变换，并将其应用于图像。
    Args:
        img: Image to be applied transformation.
        transforms (list): List of transformations to be chosen from to apply.

    Returns:
        Transformed image.
    """
    return random.choice(transforms)(img)

def uniform_augment(img, transforms, num_ops):
    """
    为每个变换随机分配一个概率, 每个图像决定是否应用它。

    Args:
        img: Image to be applied transformation.
        transforms (list): List of transformations to be chosen from to apply.
        num_ops (int): number of transforms to sequentially aaply.

    Returns:
        Transformed image.
    """
    op_idx = np.random.choice(len(transforms), size=num_ops, replace=False)
    for idx in op_idx:
        augment_op = transforms[idx]
        pr = random.random()
        if random.random() < pr:
            img = augment_op(img.copy())
    return img

class Compose(object):
    """
    Usage Directions:
        Used to combine transforms.
    """
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, label=None):
        if label is not None:
            for t in self.transforms:
                image, label = t(image, label)
            return image, label
        else:
            for t in self.transforms:
                image = t(image)
            return image

class RandomApply():
    def __init__(self, transforms, prob=0.5):
        self.prob = prob
        self.transforms = transforms

    def __call__(self, img):
        return random_apply(img, self.transforms, self.prob)

class RandomChoice():
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        return random_choice(img, self.transforms)

class RandomOrder():
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        return random_order(img, self.transforms)

class UniformAugment():
    def __init__(self, transforms, num_ops):
        self.transforms = transforms
        self.num_ops = num_ops

    def __call__(self, img):
        return uniform_augment(img, self.transforms, self.num_ops)


import torch
import numpy as np
from pyzjr.utils.check import is_tensor, is_numpy

def hwc2chw(x):
    """
    Conversion from 'HWC' to 'CHW' format.
    Example:
        hwc_image_numpy = np.random.rand(256, 256, 3)
        chw_image_numpy = hwc2chw(hwc_image_numpy)
        hwc_image_tensor = torch.rand(256, 256, 3)
        chw_image_tensor = hwc2chw(hwc_image_tensor)
    """
    if len(x.shape) == 3:
        if is_numpy(x):
            chw = np.transpose(x, axes=[2, 0, 1])
            return chw
        elif is_tensor(x):
            chw = x.permute(2, 0, 1).contiguous()
            return chw
        else:
            raise TypeError("The input data should be a NumPy array or "
                            "PyTorch tensor, but the provided type is: {}".format(type(img)))
    else:
        raise ValueError("The input data should be three-dimensional (height x width x channel), but the "
                         "provided number of dimensions is:{}".format(len(img.shape)))

def chw2hwc(x):
    """Conversion from 'CHW' to 'HWC' format."""
    if len(x.shape) == 3:
        if is_numpy(x):
            hwc = np.transpose(x, axes=[1, 2, 0])
            return hwc
        elif is_tensor(x):
            hwc = x.permute(1, 2, 0).contiguous()
            return hwc
        else:
            raise TypeError("The input data should be a NumPy array or "
                            "PyTorch tensor, but the provided type is: {}".format(type(img)))
    else:
        raise ValueError ("The input data should be three-dimensional (channel x height x width), but the "
                          "provided number of dimensions is: {}".format(len(img.shape)))

def to_bchw(x):
    """
    Convert to 'bchw' format
    Example:
        image_tensor = torch.rand(256, 256)
        bchw_image_tensor = to_bchw(image_tensor)
        print("Original shape:", image_tensor.shape)
        print("Converted shape:", bchw_image_tensor.shape)
    """
    if len(x.shape) < 2:
        raise ValueError(f"Input size must be a two, three or four dimensional tensor. Got {tensor.shape}")

    if len(x.shape) == 2:
        if is_tensor(x):
            x = x.unsqueeze(0)
        elif is_numpy(x):
            x = np.expand_dims(x, axis=0)

    if len(x.shape) == 3:
        if is_tensor(x):
            x = x.unsqueeze(0)
        elif is_numpy(x):
            x = np.expand_dims(x, axis=0)

    if len(x.shape) > 4:
        if is_tensor(x):
            x = x.view(-1, x.shape[-3], x.shape[-2], x.shape[-1])
        elif is_numpy(x):
            x = x.reshape((-1, x.shape[-3], x.shape[-2], x.shape[-1]))
    return x

if __name__=="__main__":
    x = torch.rand(2, 4, 3, 256, 256)
    x_n = np.random.rand(2, 4, 3, 256, 256)
    c = to_bchw(x)
    c_ = to_bchw(x_n)
    print(c.shape, c_.shape)


from .clean import clear_directory, cleanup_directory, rm_makedirs
from .txtfun import generate_txt, read_txt, write_to_txt
from .tuplefun import (
    _ntuple, to_1tuple,
    to_2tuple, to_3tuple,
    to_4tuple, to_ntuple,
)
from .folder import (
    multi_makedirs, unique_makedirs, datatime_makedirs, logdir, loss_weights_dirs, timestr
)
from .file import (
    file_age,
    file_date,
    file_size,
    read_yaml,
    read_json
)
from .listfun import (
    natsorted,
    natural_key,
    list_alphabet,
    list_dirs,
    list_files,
)
from .getpath import (
    getPhotoPath,
    SearchFilePath,
    SearchFileName,
    split_path2list,
    SearchSpecificFilePath,
    getSpecificImages,
)


"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import cv2
from pylab import *
from skimage.morphology import disk
from skimage.filters import rank
from skimage import measure
from pyzjr.augmentation.mask.predeal import binarization
import pyzjr.Z as Z

def getContours(img, cThr=(50, 100), minArea=1000, filter=0, draw=True):
    """
    :param img: 输入图像
    :param cThr: 阈值
    :param minArea: 更改大小
    :param filter: 过滤
    :param draw: 绘制边缘
    :return: 返回图像轮廓
    """
    imgGray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    imgBlur = cv2.GaussianBlur(imgGray, (5, 5), 1)
    imgCanny = cv2.Canny(imgBlur, cThr[0], cThr[1])
    kernel = np.ones((5, 5))
    imgDial = cv2.dilate(imgCanny, kernel, iterations=3)
    imgThre = cv2.erode(imgDial, kernel, iterations=2)
    contours, _ = cv2.findContours(imgThre, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    finalCountours = []
    for i in contours:
        area = cv2.contourArea(i)
        if area > minArea:
            peri = cv2.arcLength(i, True)
            approx = cv2.approxPolyDP(i, 0.02 * peri, True)
            bbox = cv2.boundingRect(approx)
            if filter > 0:
                if len(approx) == filter:
                    finalCountours.append([len(approx), area, approx, bbox, i])
            else:
                finalCountours.append([len(approx), area, approx, bbox, i])
    finalCountours = sorted(finalCountours, key=lambda x: x[1], reverse=True)
    if draw:
        for con in finalCountours:
            cv2.drawContours(img, con[4], -1, (0, 0, 255), 3)
    return img, finalCountours

def SearchOutline(binary_image):
    """
    在给定图像中搜索轮廓并返回轮廓的坐标列表。
    :param binary_image: 要搜索轮廓的图像。
    :return: 包含轮廓坐标的列表，每个坐标表示裂缝的一个点，坐标格式为 [(x, y),...]。
    """
    contours = measure.find_contours(binary_image, level=128, fully_connected='low', positive_orientation='low')
    contours_xy = [np.fliplr(np.vstack(contour)).astype(np.int32) for contour in contours]
    return contours_xy

def drawOutline(blackbackground, contours, color=Z.purple, thickness=1):
    """绘制轮廓"""
    cv2.drawContours(blackbackground, contours, -1, color, thickness=thickness)

def gradientOutline(img, radius=2):
    """
    对图像进行梯度边缘检测，并返回边缘强度图像。
    
    :param img: 输入的图像(内部有二值转化)
    :param radius: 半径,默认为2。
    :return: 返回经过梯度边缘检测处理后的边缘强度图像
    """
    image = binarization(img)
    denoised = rank.median(image, disk(radius))
    gradient = rank.gradient(denoised, disk(radius))
    gradient = cv2.cvtColor(gradient, cv2.COLOR_GRAY2BGR)

    return gradient

def sort_contours(cnts, method="left-to-right"):
    # initialize the reverse flag and sort index
    reverse = False
    i = 0
    if method == "right-to-left" or method == "bottom-to-top":
        reverse = True
    if method == "top-to-bottom" or method == "bottom-to-top":
        i = 1
    boundingBoxes = [cv2.boundingRect(c) for c in cnts]
    (cnts, boundingBoxes) = zip(*sorted(zip(cnts, boundingBoxes),
                                        key=lambda b: b[1][i], reverse=reverse))
    return cnts, boundingBoxes


def label_contour(image, c, i=1, prefix="#", color=(0, 255, 0), thickness=2):
    """
    在图像上标记轮廓，并在轮廓中心显示标签。

    参数:
        image: 输入的图像（numpy.ndarray）。
        c: 轮廓点集（numpy.ndarray）。
        i: 轮廓的索引值（int）。
        prefix: 标签的前缀，默认为 "#"（str）。
        color: 轮廓的颜色，默认为绿色 (0, 255, 0)（tuple）。
        thickness: 轮廓线条的粗细，默认为 2（int）。
    返回:
        标记了轮廓和标签的图像。
    """
    M = cv2.moments(c)
    cX = int(M["m10"] / M["m00"])
    cY = int(M["m01"] / M["m00"])

    cv2.drawContours(image, [c], -1, color, thickness)
    cv2.putText(image, f"{prefix}{i}", (cX - 20, cY), cv2.FONT_HERSHEY_SIMPLEX,
                1.0, (255, 255, 255), 2)
    return image


def foreground_contour_length(binary_img, minArea=5):
    """
    计算前景的轮廓长度, 返回两个值, 每个轮廓的长度和总长度
    :param binary_img: 二值图
    :param minArea: 最小过滤面积
    """
    contours_xy = SearchOutline(binary_img)
    contour_lengths = []
    for contour in contours_xy:
        area = cv2.contourArea(contour)
        if area > minArea:
            length = cv2.arcLength(contour, False)
            contour_lengths.append(length)
    all_length = np.sum(contour_lengths)
    return contour_lengths, all_length


def check_points_in_contour(center, radius, contour):
    x, y = center
    directions = [
        (x + radius, y),  # 右
        (x - radius, y),  # 左
        (x, y + radius),  # 下
        (x, y - radius),  # 上
        (x + radius, y + radius),  # 右下
        (x + radius, y - radius),  # 右上
        (x - radius, y + radius),  # 左下
        (x - radius, y - radius),  # 左上
    ]
    for pt in directions:
        if cv2.pointPolygonTest(contour, pt, False) < 0:  # 点在轮廓外
            return False
    return True


if __name__=="__main__":
    image = np.zeros((500, 500, 3), dtype=np.uint8)
    contour = np.array([[100, 100], [400, 100], [400, 400], [100, 400]], dtype=np.int32)
    labeled_image = label_contour(image, contour, 0)
    cv2.imshow("Labeled Contour", labeled_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


"""
Copyright (c) 2023, Auorui.
All rights reserved.

Xception: Deep Learning with Depthwise Separable Convolutions
    <https://arxiv.org/pdf/1610.02357.pdf>

Blog records: https://blog.csdn.net/m0_62919535/article/details/137065439
"""
import torch
import torch.nn as nn

__all__ = ["Xception"]

class SeparableConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1, dilation=1, bias=False):
        super(SeparableConv2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size, stride, padding,
                               dilation, groups=in_channels, bias=bias)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0,
                                   dilation=1, groups=1, bias=False)

    def forward(self, x):
        x = self.conv(x)
        x = self.pointwise(x)
        return x

class EntryFlow(nn.Module):
    def __init__(self):
        super(EntryFlow, self).__init__()

        self.headconv = nn.Sequential(
            nn.Conv2d(3, 32, 3, 2, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU(inplace=True),

            nn.Conv2d(32, 64, 3, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )

        self.residual_block1 = nn.Sequential(
            SeparableConv2d(64, 128, 3, padding=1),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
            SeparableConv2d(128, 128, 3, padding=1),
            nn.BatchNorm2d(128),
            nn.MaxPool2d(3, stride=2, padding=1),
        )

        self.residual_block2 = nn.Sequential(
            nn.ReLU(inplace=True),
            SeparableConv2d(128, 256, 3, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            SeparableConv2d(256, 256, 3, padding=1),
            nn.BatchNorm2d(256),
            nn.MaxPool2d(3, stride=2, padding=1)
        )

        self.residual_block3 = nn.Sequential(
            nn.ReLU(inplace=True),
            SeparableConv2d(256, 728, 3, padding=1),
            nn.BatchNorm2d(728),
            nn.ReLU(inplace=True),
            SeparableConv2d(728, 728, 3, padding=1),
            nn.BatchNorm2d(728),
            nn.MaxPool2d(3, stride=2, padding=1)
        )

    def shortcut(self, inp, oup):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 1, 2, bias=False),
            nn.BatchNorm2d(oup)
            )

    def forward(self, x):
        x = self.headconv(x)
        residual = self.residual_block1(x)
        shortcut_block1 = self.shortcut(64, 128)
        x = residual + shortcut_block1(x)
        residual = self.residual_block2(x)
        shortcut_block2 = self.shortcut(128, 256)
        x = residual + shortcut_block2(x)
        residual = self.residual_block3(x)
        shortcut_block3 = self.shortcut(256, 728)
        x = residual + shortcut_block3(x)

        return x

class MiddleFlow(nn.Module):
    def __init__(self):
        super(MiddleFlow, self).__init__()

        self.shortcut = nn.Sequential()
        self.conv1 = nn.Sequential(
            nn.ReLU(inplace=True),
            SeparableConv2d(728, 728, 3, padding=1),
            nn.BatchNorm2d(728),

            nn.ReLU(inplace=True),
            SeparableConv2d(728, 728, 3, padding=1),
            nn.BatchNorm2d(728),

            nn.ReLU(inplace=True),
            SeparableConv2d(728, 728, 3, padding=1),
            nn.BatchNorm2d(728)
        )

    def forward(self, x):
        residual = self.conv1(x)
        input = self.shortcut(x)
        return input + residual

class ExitFlow(nn.Module):
    def __init__(self):
        super(ExitFlow, self).__init__()

        self.residual_with_exit = nn.Sequential(
            nn.ReLU(inplace=True),
            SeparableConv2d(728, 728, 3, padding=1),
            nn.BatchNorm2d(728),
            nn.ReLU(inplace=True),
            SeparableConv2d(728, 1024, 3, padding=1),
            nn.BatchNorm2d(1024),
            nn.MaxPool2d(3, stride=2, padding=1)
        )

        self.endconv = nn.Sequential(
            SeparableConv2d(1024, 1536, 3, 1, 1),
            nn.BatchNorm2d(1536),
            nn.ReLU(inplace=True),

            SeparableConv2d(1536, 2048, 3, 1, 1),
            nn.BatchNorm2d(2048),
            nn.ReLU(inplace=True),

            nn.AdaptiveAvgPool2d((1, 1)),
        )

    def shortcut(self, inp, oup):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 1, 2, bias=False),
            nn.BatchNorm2d(oup)
        )

    def forward(self, x):
        residual = self.residual_with_exit(x)
        shortcut_block = self.shortcut(728, 1024)
        output = residual + shortcut_block(x)
        return self.endconv(output)


class Xception(nn.Module):
    def __init__(self, num_classes=1000):
        super().__init__()
        self.num_classes = num_classes

        self.entry_flow = EntryFlow()
        self.middle_flow = MiddleFlow()
        self.exit_flow = ExitFlow()
        self.fc = nn.Linear(2048, num_classes)

    def forward(self, x):
        x = self.entry_flow(x)
        for i in range(8):
            # Repeated 8 times
            x = self.middle_flow(x)
        x = self.exit_flow(x)
        x = x.view(x.size(0), -1)
        out = self.fc(x)

        return out

if __name__=='__main__':
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = Xception(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Xception Total params: 19,838,076


from datetime import datetime
from pathlib import Path
import yaml
import json

def read_yaml(path, name=None):
    with open(path) as file:
        data = yaml.load(file, Loader=yaml.FullLoader)
    value = data[name] if name is not None and name in data else data

    return value

def read_json(path, name=None):
    with open(path, 'r') as file:
        data = json.load(file)
    value = data[name] if name is not None and name in data else data

    return value

def file_age(path, detail=False):
    """Returns the number of days since the last file update."""
    dt = (datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime))  # delta
    se = 0
    if detail:
        se = dt.seconds / 86400
    return f"{dt.days + se} days"

def file_date(path=__file__):
    """Return readable file modification date,i.e:'2021-3-26'."""
    t = datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'

def file_size(path):
    """Return file/dir size (MB)."""
    if isinstance(path, (str, Path)):
        mb = 1 << 20  # bytes -> MiB (1024 ** 2)
        path = Path(path)
        if path.is_file():
            size_mb = path.stat().st_size / mb
            return f"{size_mb:.5f} MB"
        elif path.is_dir():
            total_size = sum(f.stat().st_size for f in path.glob('**/*') if f.is_file())
            size_mb = total_size / mb
            return f"{size_mb:.5f} MB"
    return "0.0 MB"



if __name__ == "__main__":
    # 测试 file_age
    current_file_path = Path(__file__)
    age = file_age(current_file_path)
    print(f"The file {current_file_path} is {age} old.")
    detail_age = file_age(current_file_path, detail=True)
    print(f"The file {current_file_path} is {detail_age} old (with detail).")

    # 测试 file_date
    date = file_date(current_file_path)
    print(f"The file {current_file_path} was modified on {date}.")

    # 测试 file_size (文件)
    file_to_check = Path(__file__)
    size = file_size(file_to_check)
    print(f"The file {file_to_check} is {size} large.")

    # 测试 file_size (目录)
    dir_to_check = Path('.')
    dir_size = file_size(dir_to_check)
    print(f"The directory {dir_to_check} is {dir_size} large.")



import os
import pyzjr.Z as Z
from pathlib import Path
from pyzjr.data.utils.listfun import natsorted

def getPhotoPath(path, debug=False):
    """
    :param path: 文件夹路径
    :param debug: 开启打印文件名错误的名字
    :return: 包含图片路径的列表
    """
    imgfile = []
    allfile = []
    file_list = os.listdir(path)
    for i in file_list:
        if debug:
            if i[0] in ['n', 't', 'r', 'b', 'f'] or i[0].isdigit():
                print(f"File name error occurred at the beginning of {i}!")
        newph = os.path.join(path, i).replace("\\", "/")
        allfile.append(newph)
        _, file_ext = os.path.splitext(newph)
        if file_ext[1:] in Z.IMG_FORMATS:
            imgfile.append(newph)

    return natsorted(imgfile), natsorted(allfile)

def SearchFilePath(filedir, file_ext='.png'):
    """What is returned is a list that includes all paths under the target path that match the suffix."""
    search_file_path = []
    for root, dirs, files in os.walk(filedir):
        for filespath in files:
            if str(filespath).endswith(file_ext):
                search_file_path.append(os.path.join(root, filespath))
    return natsorted(search_file_path)

def SearchFileName(target_path, file_ext='.png'):
    """Only search for file names in the appropriate format in the target folder"""
    all_files = os.listdir(target_path)
    png_files = [file for file in all_files if file.lower().endswith(file_ext)]
    sorted_png_files = natsorted(png_files)
    return sorted_png_files

def split_path2list(path_str):
    """
    path_list = split_path2list('D:\PythonProject\MB_TaylorFormer\DehazeFormer\data\rshazy\test\GT\220.png')
    Return:
        ['D:\\', 'PythonProject', 'MB_TaylorFormer', 'DehazeFormer', 'data', 'rshazy', 'test', 'GT', '220.png']
    """
    path = Path(path_str)
    path_parts = path.parts
    return list(path_parts)

def getSpecificImages(basePath, contains=None):
    # return the set of files that are valid
    return list(SearchSpecificFilePath(basePath, validExts=Z.IMG_FORMATS, contains=contains))

def SearchSpecificFilePath(basePath, validExts=None, contains=None):
    # loop over the directory structure
    for (rootDir, dirNames, filenames) in os.walk(basePath):
        for filename in filenames:
            if contains is not None and filename.find(contains) == -1:
                continue
            ext = filename[filename.rfind("."):].lower()

            if validExts is None or ext.endswith(validExts):
                imagePath = os.path.join(rootDir, filename)
                yield imagePath



if __name__ == "__main__":
    path = r"E:\PythonProject\pyzjrPyPi"
    print(SearchFilePath(path))
    print(getSpecificImages(path))


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used for color space conversion.
"""
import colorsys
import cv2
import numpy as np
from PIL import Image
from pyzjr.utils.check import is_numpy, is_pil

def pil2cv(pil_image):
    """将PIL图像转换为OpenCV图像"""
    if pil_image.mode == 'L':
        open_cv_image = np.array(pil_image)
    else:
        open_cv_image = cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)
    return open_cv_image

def cv2pil(cv_image):
    """将OpenCV图像转换为PIL图像"""
    if cv_image.ndim == 2:
        pil_image = Image.fromarray(cv_image, mode='L')
    else:
        pil_image = Image.fromarray(cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB))
    return pil_image

def to_gray(image):
    if is_pil(image):
        return image.convert('L')
    elif is_numpy(image):
        return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")

def bgr2rgb(image):
    if is_pil(image):
        image_np = np.asarray(image)
        image_rgb = image_np[:, :, ::-1]
        return Image.fromarray(image_rgb).convert("RGB")
    elif is_numpy(image):
        return image[:, :, ::-1]
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")

def rgb2bgr(image):
    if is_pil(image):
        image_np = np.asarray(image)
        image_bgr = image_np[:, :, ::-1]
        return Image.fromarray(image_bgr)
    elif is_numpy(image):
        return image[:, :, ::-1]
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")

def to_hsv(image):
    if is_pil(image):
        image_np = np.asarray(image)
        image_hsv = cv2.cvtColor(image_np, cv2.COLOR_RGB2HSV)
        return Image.fromarray(image_hsv)
    elif is_numpy(image):
        return cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")

def hsv2rgb(image):
    if is_pil(image):
        image_np = np.asarray(image)
        image_rgb = cv2.cvtColor(image_np, cv2.COLOR_HSV2RGB)
        return Image.fromarray(image_rgb)
    elif is_numpy(image):
        return cv2.cvtColor(image, cv2.COLOR_HSV2RGB)
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")

def hsv2bgr(image):
    if is_pil(image):
        image_np = np.asarray(image)
        image_bgr = cv2.cvtColor(image_np, cv2.COLOR_HSV2BGR)
        return Image.fromarray(image_bgr)
    elif is_numpy(image):
        return cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
    else:
        raise TypeError("Unsupported image type. Expected PIL Image or OpenCV (NumPy array).")


def create_palette(num_classes):
    """
    创建一个颜色调色板，用于可视化不同类别的颜色。
    - 如果num_classes <= 21，使用 VOC 数据集的预定义颜色。
    - 如果num_classes > 21，使用 HSV 颜色空间生成均匀分布的颜色。
    """
    if num_classes <= 21:
        colors = [[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0],
                  [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
                  [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
                  [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
                  [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
                  [0, 64, 128]]
    else:
        hsv_tuples = [(x / num_classes, 1., 1.) for x in range(num_classes)]
        colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))
    return colors

if __name__=="__main__":
    image_path = r"E:\PythonProject\pyzjrPyPi\pyzjr\test.png"
    image_pil = Image.open(image_path)
    # gray_pil = to_gray(image_pil)
    image_pil = rgb2bgr(image_pil)
    image_pil = bgr2rgb(image_pil)
    image_pil = to_hsv(image_pil)
    image_pil = hsv2rgb(image_pil)
    image_pil.show()

    image_cv = cv2.imread(image_path)
    image_cv = bgr2rgb(image_cv)
    image_cv = rgb2bgr(image_cv)
    image_cv = to_hsv(image_cv)
    image_cv = hsv2bgr(image_cv)
    cv2.imshow('Gray Image', image_cv)
    # gray_cv = to_gray(image_cv)
    # cv2.imshow('Gray Image', gray_cv)
    cv2.waitKey(0)
    cv2.destroyAllWindows()




from .se_module import SqueezeExcitation, SEAttention, EffectiveSEAttention
from .bam_module import BAMAttention
from .cbam_module import CBAMAttention, LightCBAMAttention
from .cot_module import CoTAttention


"""
Copyright (c) 2022, Auorui.
All rights reserved.
"""
import cv2
import torch
import numpy as np
_imshow = cv2.imshow  # copy to avoid recursion errors
from pathlib import Path
from pyzjr.utils.check import is_file, is_numpy, is_list, is_pil, is_tensor
from pyzjr.augmentation.augments import resizepad
from pyzjr.utils.randfun import randstring

def read_gray(filename):
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    return cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_GRAYSCALE)

def read_bgr(filename):
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    return cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_COLOR)

def read_rgb(filename):
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    return cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_COLOR)[:, :, ::-1]

def read_tensor(filename, target_shape, device=None, pad_color=None):
    if device is None:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    rgb_image = read_rgb(filename)
    if pad_color:
        image = resizepad(rgb_image, target_shape, pad_color)
    else:
        image = cv2.resize(rgb_image, target_shape)
    image = image.astype(np.float32) / 255.0
    image_tensor = torch.from_numpy(image).permute(2, 0, 1).unsqueeze(0)
    return image_tensor.to(device)

def read_image(filename, flags='rgb', target_shape=None, pad_color=None):
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    image = None
    if flags in ['bgr', 'BGR']:
        image = cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_COLOR)
    elif flags in ['rgb', 'RGB', 'torch', 'tensor', 'Tensor']:
        image = cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_COLOR)[:, :, ::-1]
    elif flags in ['alpha', 'Alpha']:
        image = cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_UNCHANGED)
    elif flags in ['gray', 'Gray']:
        image = cv2.imdecode(np.fromfile(filename, np.uint8), cv2.IMREAD_GRAYSCALE)
    if target_shape:
        if pad_color:
            image = resizepad(image, target_shape, pad_color)
        else:
            image = cv2.resize(image, target_shape)

    if flags in ['torch', 'tensor', 'Tensor']:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        image = image.astype(np.float32) / 255.0
        image = torch.from_numpy(image).permute(2, 0, 1).unsqueeze(0)
        image = image.to(device)
    return image

def imwrite(filename: str, image, params=None):
    """Write the image to a file."""
    if is_tensor(image):
        image = image.detach().cpu().squeeze(0)
        image = image.permute(1, 2, 0).numpy()
        image = (image - image.min()) / (image.max() - image.min())
        image = (image[:, :, ::-1] * 255).astype(np.uint8)
    try:
        cv2.imencode(Path(filename).suffix, image, params)[1].tofile(filename)
        return True
    except Exception:
        return False

def display(image, winname=None, scale=1.):
    """Displays an image in the specified window."""
    if winname is None:
        winname = randstring(5)
    if is_numpy(image):
        height, width = image.shape[:2]
        new_width = int(width * scale)
        new_height = int(height * scale)
        image = cv2.resize(image, (new_width, new_height))
    elif is_list(image):
        image = StackedImagesV1(scale, image)
    _imshow(winname.encode('unicode_escape').decode(), image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()

def StackedImagesV1(scale, imgArray):
    """
    Display Images According to List Structure

    :param scale: The scale of the images, where 1 represents the original size.
    :param imgArray: A list of images representing the arrangement in rows and columns.
    :return: A generated image that displays the images in the order specified by the input list, arranged in a grid.
    """
    rows = len(imgArray)
    cols = len(imgArray[0])
    rowsAvailable = isinstance(imgArray[0], list)
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]
    if rowsAvailable:
        for x in range(0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape[:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank] * rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)
    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2:
                imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor = np.hstack(imgArray)
        ver = hor
    return ver

def StackedImagesV2(scale, imgList, cols):
    """
    Combine multiple images into a single display within a single window
    :param scale: The scaling factor for the images, where a value greater than 1 indicates enlargement and a value less than 1 indicates reduction.
    :param imgList: A list of images to be combined.
    :param cols: The number of images to display per row.
    :return: The combined image.
    """
    totalImages = len(imgList)
    rows = totalImages // cols if totalImages // cols * cols == totalImages else totalImages // cols + 1
    blankImages = cols * rows - totalImages

    width = imgList[0].shape[1]
    height = imgList[0].shape[0]
    imgBlank = np.zeros((height, width, 3), np.uint8)
    imgList.extend([imgBlank] * blankImages)
    for i in range(cols * rows):
        imgList[i] = cv2.resize(imgList[i], (0, 0), None, scale, scale)
        if len(imgList[i].shape) == 2:
            imgList[i] = cv2.cvtColor(imgList[i], cv2.COLOR_GRAY2BGR)
    hor = [imgBlank] * rows
    for y in range(rows):
        line = []
        for x in range(cols):
            line.append(imgList[y * cols + x])
        hor[y] = np.hstack(line)
    ver = np.vstack(hor)
    return ver

def imattributes(image):
    """Retrieve image attributes"""
    if is_numpy(image):
        imshape = image.shape
        size = imshape[:2]
        dtype = image.dtype
        height, width = size

        if len(imshape) == 2:
            depth = image.itemsize * 8
        else:
            depth = image.itemsize * 8 * imshape[2]

        return {
            "shape": imshape,
            "size": size,
            "height": height,
            "width": width,
            "dtype": dtype,
            "depth": depth,
            "source": "NumPy"
        }
    elif is_pil(image):
        width, height = image.size
        mode = image.mode

        if mode == 'L':  # Grayscale
            depth = 8
        elif mode in ('RGB', 'RGBA'):
            depth = 8 * 3 if mode == 'RGB' else 8 * 4
        elif mode in ('I', 'F'):  # Integer or float (single channel)
            depth = image.getbands()[0].size * 8
        else:
            depth = "Unknown"  # Other modes are not handled here

        return {
            "shape": (height, width, len(image.getbands())),  # Simulate a NumPy shape
            "size": (width, height),
            "height": height,
            "width": width,
            "dtype": "PIL mode " + mode,  # PIL doesn't have a direct dtype equivalent
            "depth": depth,
            "source": "PIL"
        }
    else:
        return "Input is not a NumPy array or PIL Image."

if __name__=="__main__":
    image_path = r"E:\PythonProject\pyzjrPyPi\pyzjr\test.png"
    image = read_image(image_path, flags='rgb')
    display(image)
    print(image.shape)


"""
CBAM (Convolutional Block Attention Module)
Original paper addresshttps: https://arxiv.org/pdf/1807.06521.pdf
Blog records: https://blog.csdn.net/m0_62919535/article/details/136334691
Time: 2024-02-28
"""
import torch
from torch import nn
import torch.nn.functional as F

class ChannelAttention(nn.Module):
    def __init__(self, in_planes, reduction_ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        # shared MLP
        self.mlp = nn.Sequential(
            nn.Conv2d(in_planes, in_planes // reduction_ratio, 1, bias=False),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_planes // reduction_ratio, in_planes, 1, bias=False)
        )
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.mlp(self.avg_pool(x))
        max_out = self.mlp(self.max_pool(x))
        out = avg_out + max_out
        return self.sigmoid(out)


class SpatialAttention(nn.Module):
    def __init__(self, kernel_size=7, padding=3):
        super(SpatialAttention, self).__init__()
        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)


class CBAMAttention(nn.Module):
    def __init__(self, in_planes, reduction_ratio=16, kernel_size=7):
        super(CBAMAttention, self).__init__()
        self.ca = ChannelAttention(in_planes, reduction_ratio)
        self.sa = SpatialAttention(kernel_size)

    def forward(self, x):
        out = x * self.ca(x)
        result = out * self.sa(out)
        return result


class LightChannelAttention(ChannelAttention):
    """An experimental 'lightweight' that sums avg + max pool first"""
    def __init__(self, channels, reduction=16):
        super(LightChannelAttention, self).__init__(channels, reduction)

    def forward(self, x):
        x_pool = 0.5 * x.mean((2, 3), keepdim=True) + 0.5 * F.adaptive_max_pool2d(x, 1)
        x_attn = self.mlp(x_pool)
        return x * x_attn.sigmoid()

class LightSpatialAttention(nn.Module):
    """An experimental 'lightweight' variant that sums avg_pool and max_pool results."""
    def __init__(self, kernel_size=7, padding=3):
        super(LightSpatialAttention, self).__init__()
        self.conv1 = nn.Conv2d(1, 1, kernel_size, padding=padding, bias=False)

    def forward(self, x):
        x_avg = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x_attn = 0.5 * x_avg + 0.5 * max_out
        x_attn = self.conv1(x_attn)
        return x * x_attn.sigmoid()

class LightCBAMAttention(nn.Module):
    def __init__(self, in_planes, reduction_ratio=16, kernel_size=7):
        super(LightCBAMAttention, self).__init__()
        self.channel = LightChannelAttention(in_planes, reduction_ratio)
        self.spatial = LightSpatialAttention(kernel_size)

    def forward(self, x):
        x = self.channel(x)
        x = self.spatial(x)
        return x

if __name__ == '__main__':
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(1, 16, 8, 8).to(device)
    net = LightCBAMAttention(16)
    net = net.to(device)
    out = net(input)
    print(out.shape)
    torchsummary.summary(net, input_size=(16, 8, 8))
    # CBAMAttention Total params: 162
    # LightCBAMAttention Trainable params: 81


"""
Copyright (c) 2024, Auorui.
All rights reserved.

This module is used as the base class for homemade data loaders.
"""
import cv2
import os
import random
import numpy as np
from torch.utils.data import Dataset
from pyzjr.utils.check import is_tensor, is_numpy
from pyzjr.augmentation.augments import resizepad

class BaseDataset(Dataset):
    """A simple dataset class with the same functionality as torch.utils.data.Dataset"""
    def __init__(self):
        pass

    def __getitem__(self, idx):
        pass

    def __len__(self):
        pass

    def read_image(self, filename, target_shape=None, to_rgb=True, normalize=True, pad_color=(128, 128, 128)):
        if not os.path.exists(filename):
            raise FileNotFoundError(f"File not found: {filename}")
        img = cv2.imread(filename, cv2.IMREAD_COLOR)
        if img is None:
            raise ValueError(f"Failed to read image: {filename}")
        if target_shape is not None:
            img = resizepad(img, target_shape, pad_color=pad_color)
        img = img[:, :, ::-1] if to_rgb else img
        return self.preprocess_input(img) if normalize else img

    def colormap2label(self, image, colormap, dtype="int64"):
        colormap_label = np.zeros(256 ** 3, dtype=np.int64)
        image = np.array(image, dtype=dtype)
        for i, color_map in enumerate(colormap):
            colormap_label[(color_map[0] * 256 + color_map[1]) * 256 + color_map[2]] = i
        idx = ((image[:, :, 0] * 256 + image[:, :, 1]) * 256
               + image[:, :, 2])
        return colormap_label[idx]

    def rand(self, a=0., b=1.):
        return random.uniform(a, b)

    def shuffle(self, x):
        random.seed(0)
        return random.sample(x, len(x))

    def multi_makedirs(self, *args):
        for path in args:
            if not os.path.exists(path):
                os.makedirs(path)

    def preprocess_input(self, image, mean=None, std=None):
        image = np.asarray(image)
        if np.max(image) > 1:
            image = image / 255.0

        if mean is not None:
            mean = np.array(mean)
            image = image - mean

        if std is not None:
            std = np.array(std)
            image = image / std

        return image.astype(np.float32)

    def read_txt(self, txt_path):
        path_list = []
        if isinstance(txt_path, str):  # Check if txt_path is a string
            with open(txt_path, 'r') as file:
                lines = file.readlines()
                path_list.extend(line.strip() for line in lines)
        return path_list

    def hwc2chw(self, image):
        if len(image.shape) == 3:
            if is_numpy(image):
                return np.transpose(image, axes=[2, 0, 1])
            elif is_tensor(image):
                return image.permute(2, 0, 1).contiguous()
            else:
                raise TypeError("The input data should be a NumPy array or "
                                "PyTorch tensor, but the provided type is: {}".format(type(image)))
        else:
            raise ValueError("The input data should be three-dimensional (height x width x channel), but the "
                             "provided number of dimensions is:{}".format(len(image.shape)))

    def auguments(self, imglists, target_shape=None, prob=.5):
        if target_shape is not None:
            H, W, _ = imglists[0].shape
            Hc, Wc = target_shape
            Hs = random.randint(0, H - Hc)
            Ws = random.randint(0, W - Wc)
            for i in range(len(imglists)):
                imglists[i] = imglists[i][Hs:(Hs + Hc), Ws:(Ws + Wc), :]

        # horizontal flip
        if random.random() > (1 - prob):
            for i in range(len(imglists)):
                imglists[i] = np.flip(imglists[i], axis=1).copy()

        r = random.randint(0, 3)
        for i in range(len(imglists)):
            imglists[i] = np.rot90(imglists[i], r, (0, 1)).copy()

        return imglists

    def align(self, imglists, target_shape):
        H, W, _ = imglists[0].shape
        Hc, Wc = target_shape
        Hs = (H - Hc) // 2
        Ws = (W - Wc) // 2
        for i in range(len(imglists)):
            imglists[i] = imglists[i][Hs:(Hs + Hc), Ws:(Ws + Wc), :]

        return imglists

    def disable_cv2_multithreading(self):
        """
        Disable OpenCV's multithreading and OpenCL usage for consistent behavior and performance.
        """
        cv2.setNumThreads(0)
        cv2.ocl.setUseOpenCL(False)

if __name__=="__main__":
    datasets = BaseDataset()
    x = ['1', '2', '3', '4', '5']
    x = datasets.shuffle(x)
    print(x)


"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import os
import cv2
import sys

class VideoCap():
    """
    Customized Python video reading class
    Examples:
    ```
        Vcap = VideoCap(mode=0)
        while True:
            img = Vcap.read()
            Vcap.show("ss", img)
    ```
    """
    def __init__(self, mode=0, width=640, height=480, light=150):
        self.cap = cv2.VideoCapture(mode)
        self.cap.set(3, width)
        self.cap.set(4, height)
        self.cap.set(10, light)
        self.start_number = 0

    def read(self, flip=None):
        """
        :param flip: -1: Horizontal and vertical directions,
                      0: along the y-axis, vertical,
                      1: along the x-axis, horizontal
        """
        _, img = self.cap.read()
        if flip is not None:
            assert flip in [-1, 0, 1], f"VideoCap: The 'flip' parameter must be -1, 0, or 1."
            img = cv2.flip(img, flip)
        return img

    def free(self):
        """
        Release camera
        """
        self.cap.release()
        cv2.destroyAllWindows()

    def show(self, winname, src, base_name: str = './result.png', end_k=27,
             save_k=ord('s'), delay_t=1, extend_num=3):
        """
        Window display. Press 's' to save, 'Esc' to end
        """
        image_path, ext = os.path.splitext(base_name)
        os.makedirs(os.path.dirname(base_name), exist_ok=True)
        if src is not None:
            cv2.imshow(winname, src)
            k = cv2.waitKey(delay_t) & 0xFF
            if k == end_k:
                self.free()
                sys.exit(0)
            elif k == save_k:
                self.start_number += 1
                file_number = str(self.start_number).zfill(extend_num)
                file_path = f"{image_path}_{file_number}{ext}"
                print(f"{self.start_number}  Image saved to {file_path}")
                cv2.imwrite(file_path, src)



if __name__=="__main__":
    Vcap = VideoCap(mode=0)
    while True:
        img = Vcap.read()
        Vcap.show("ss", img)


from .conv_norm_act import (
    ConvNormAct,
    ConvNorm,
    NormAct,
    ConvAct,
    conv3x3,
    conv1x1,
    ConvBnReLU, ConvBNReLU,
    ConvBn, ConvBN,
    BnReLU, BNReLU,
    ConvReLU
)

from .act import (
    ArgMax,
    Clamp,
    Activation
)

from .drop import (
    DropPath,
    Dropout,
    MultiSampleDropout,
    DropConnect,
    Standout,
    GaussianDropout
)

from .initer import (
    init_weights_complex,
    init_weights_simply,
    official_init,
    xavier_init,
    normal_init,
    uniform_init,
    trunc_normal_init,
)

from .useful_block import (
    ResBasicBlock,
    ResBottleneck,
    FireModule,
    DenseBlock,
    MobileInvertedResidual,
    GhostModule,
    GhostBottleneck
)


"""
Copyright (c) 2024, Auorui.
All rights reserved.

The Torch implementation of average pooling and maximum pooling has been compared with the official Torch implementation
Time: 2024-01-22  17:28
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from pyzjr.data.utils.tuplefun import to_2tuple, to_4tuple

class MaxPool2d(nn.Module):
    """
    定义MaxPool2d类，结合了普通池化和矢量化池化的功能
    池化层计算公式:
        output_size = [(input_size−kernel_size) // stride + 1]
    """
    def __init__(self, kernel_size, stride, is_vectoring=True):
        super(MaxPool2d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.is_vectoring = is_vectoring

    def max_pool2d_traditional(self, input_tensor):
        batch_size, channels, height, width = input_tensor.size()
        output_height = (height - self.kernel_size) // self.stride + 1
        output_width = (width - self.kernel_size) // self.stride + 1
        output_tensor = torch.zeros(batch_size, channels, output_height, output_width)

        for i in range(output_height):
            for j in range(output_width):
                window = input_tensor[:, :,
                         i * self.stride: i * self.stride + self.kernel_size, j * self.stride: j * self.stride + self.kernel_size]
                output_tensor[:, :, i, j] = torch.max(window.reshape(batch_size, channels, -1), dim=2)[0]
        return output_tensor

    def max_pool2d_vectorized(self, input_tensor):
        unfolded = input_tensor.unfold(2, self.kernel_size, self.stride).unfold(3, self.kernel_size, self.stride)
        max_values, _ = unfolded.max(dim=-1)
        max_values, _ = max_values.max(dim=-1)
        return max_values

    def forward(self, input_tensor):
        if self.is_vectoring:
            return self.max_pool2d_vectorized(input_tensor)
        else:
            return self.max_pool2d_traditional(input_tensor)


class AvgPool2d(nn.Module):
    """
    定义AvgPool2d类，结合了普通池化和矢量化池化的功能
    池化层计算公式:
        output_size = [(input_size−kernel_size) // stride + 1]
    """
    def __init__(self, kernel_size, stride, is_vectoring=True):
        super(AvgPool2d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.is_vectoring = is_vectoring

    def avg_pool2d_traditional(self, input_tensor):
        batch_size, channels, height, width = input_tensor.size()
        output_height = (height - self.kernel_size) // self.stride + 1
        output_width = (width - self.kernel_size) // self.stride + 1
        output_tensor = torch.zeros(batch_size, channels, output_height, output_width)

        for i in range(output_height):
            for j in range(output_width):
                window = input_tensor[:, :,
                         i * self.stride: i * self.stride + self.kernel_size, j * self.stride: j * self.stride + self.kernel_size]
                output_tensor[:, :, i, j] = torch.mean(window.reshape(batch_size, channels, -1), dim=2)
        return output_tensor

    def avg_pool2d_vectorized(self, input_tensor):
        unfolded = input_tensor.unfold(2, self.kernel_size, self.stride).unfold(3, self.kernel_size, self.stride)
        avg_values = unfolded.mean(dim=-1)
        avg_values = avg_values.mean(dim=-1)
        return avg_values

    def forward(self, input_tensor):
        if self.is_vectoring:
            return self.avg_pool2d_vectorized(input_tensor)
        else:
            return self.avg_pool2d_traditional(input_tensor)

class StridedPool2d(nn.Module):
    """
    实现一个跨步卷积层Strided Convolution, 本质上可以实现类似于池化操作的效果。
    通过步幅stride大于 1 的卷积操作来实现空间分辨率的降低。
    """
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=2, is_relu=True):
        super(StridedPool2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=1)
        self.relu = nn.ReLU(inplace=True)
        self.is_relu = is_relu

    def forward(self, x):
        x = self.conv(x)
        if self.is_relu:
            x = self.relu(x)
        return x

class MedianPool2d(nn.Module):
    """ Median pool (usable as median filter when stride=1) module.
    Hacked together by / Copyright 2020 Ross Wightman
    Currently, it is rarely used in CNN, and the code is taken
    from the project pytorch-image-models

    Args:
         kernel_size: size of pooling kernel, int or 2-tuple
         stride: pool stride, int or 2-tuple
         padding: pool padding, int or 4-tuple (l, r, t, b) as in pytorch F.pad
         same: override padding and enforce same padding, boolean
    """

    def __init__(self, kernel_size=3, stride=1, padding=0, same=False):
        super(MedianPool2d, self).__init__()
        self.k = to_2tuple(kernel_size)
        self.stride = to_2tuple(stride)
        self.padding = to_4tuple(padding)  # convert to l, r, t, b
        self.same = same

    def _padding(self, x):
        if self.same:
            ih, iw = x.size()[2:]
            if ih % self.stride[0] == 0:
                ph = max(self.k[0] - self.stride[0], 0)
            else:
                ph = max(self.k[0] - (ih % self.stride[0]), 0)
            if iw % self.stride[1] == 0:
                pw = max(self.k[1] - self.stride[1], 0)
            else:
                pw = max(self.k[1] - (iw % self.stride[1]), 0)
            pl = pw // 2
            pr = pw - pl
            pt = ph // 2
            pb = ph - pt
            padding = (pl, pr, pt, pb)
        else:
            padding = self.padding
        return padding

    def forward(self, x):
        x = F.pad(x, self._padding(x), mode='reflect')
        x = x.unfold(2, self.k[0], self.stride[0]).unfold(3, self.k[1], self.stride[1])
        x = x.contiguous().view(x.size()[:4] + (-1,)).median(dim=-1)[0]
        return x

if __name__=="__main__":
    # MaxPool2d与AvgPool2d手写测试实验
    # input_data = torch.rand((1, 3, 3, 3))
    # input_data = torch.Tensor([[[[0.3939, 0.8964, 0.3681],
    #                            [0.5134, 0.3780, 0.0047],
    #                            [0.0681, 0.0989, 0.5962]],
    #                           [[0.7954, 0.4811, 0.3329],
    #                            [0.8804, 0.3986, 0.3561],
    #                            [0.2797, 0.3672, 0.6508]],
    #                           [[0.6309, 0.1340, 0.0564],
    #                            [0.3101, 0.9927, 0.5554],
    #                            [0.0947, 0.2305, 0.8299]]]])
    #
    # print(input_data.shape)
    # is_vectoring = False
    # kernel_size = 3
    # stride = 2
    # MaxPool2d1 = nn.MaxPool2d(kernel_size, stride)
    # output_data_with_torch_max = MaxPool2d1(input_data)
    # AvgPool2d1 = nn.AvgPool2d(kernel_size, stride)
    # output_data_with_torch_avg = AvgPool2d1(input_data)
    # AvgPool2d2 = AvgPool2d(kernel_size, stride, is_vectoring)
    # output_data_with_torch_Avg = AvgPool2d2(input_data)
    # MaxPool2d2 = MaxPool2d(kernel_size, stride, is_vectoring)
    # output_data_with_torch_Max = MaxPool2d2(input_data)
    # # output_data_with_max = max_pool2d(input_data, kernel_size, stride)
    # # output_data_with_avg = avg_pool2d(input_data, kernel_size, stride)
    #
    # print("\ntorch.nn pooling Output:")
    # print(output_data_with_torch_max,"\n",output_data_with_torch_max.size())
    # print(output_data_with_torch_avg,"\n",output_data_with_torch_avg.size())
    # print("\npooling Output:")
    # print(output_data_with_torch_Max,"\n",output_data_with_torch_Max.size())
    # print(output_data_with_torch_Avg,"\n",output_data_with_torch_Avg.size())
    # # 直接使用bool方法判断会因为浮点数的原因出现偏差
    # print(torch.allclose(output_data_with_torch_max,output_data_with_torch_Max))
    # print(torch.allclose(output_data_with_torch_avg,output_data_with_torch_Avg))
    # tensor([[[[0.8964]],       # output_data_with_max
    #          [[0.8804]],
    #          [[0.9927]]]])
    # tensor([[[[0.3686]],       # output_data_with_avg
    #           [[0.5047]],
    #           [[0.4261]]]])
    #
    input_data = torch.rand((1, 3, 64, 64))
    strided_conv = StridedPool2d(3, 64)
    output_data = strided_conv(input_data)
    print("Input shape:", input_data.shape)
    print("Output shape:", output_data.shape)


import cv2
import numpy as np
from skimage.morphology import medial_axis
from scipy.ndimage import distance_transform_edt


def cv_distance(image):
    """Using distance transformation functions in OpenCV"""
    dist_transform = cv2.distanceTransform(image, cv2.DIST_L2, 3)
    return dist_transform

def chamfer(image, weight=None):
    """Chamfer distance transformation"""
    if weight is None:
        weights = np.array([[1, 1, 1],
                            [1, 0, 1],
                            [1, 1, 1]], dtype=np.uint8)
    else:
        weights = weight
    dist_transform = cv2.filter2D(image, cv2.CV_32F, weights)
    return dist_transform

def fast_marching(image):
    """Fast-Marching Distance Transform"""
    _, medial_axis_image = medial_axis(image, return_distance=True)   # 使用medial_axis函数计算中轴线
    dist_transform = distance_transform_edt(medial_axis_image)
    return dist_transform

if __name__=="__main__":
    image = np.zeros((10, 10), dtype=np.uint8)
    image[2:8, 2:8] = 255
    print(image)

    dist_transform = cv_distance(image)
    chamfer_dist_transform = chamfer(image)
    fast_marching_dist_transform = fast_marching(image)
    # 距离变换结果
    print("Distance Transform:")
    print(dist_transform)
    print("\nChamfer Distance Transform:")
    print(chamfer_dist_transform)
    print("\nFast-Marching Distance Transform:")
    print(fast_marching_dist_transform)


import cv2
import numpy as np
from skimage.morphology import skeletonize
from skimage import morphology, measure

from pyzjr.visualize.plot.text import PutBoxText
from pyzjr.augmentation.mask.predeal import binarization, bool2mask

# 裂缝的类型
class CrackType():
    """直方图投影法推断裂缝类型, 能区分线性裂缝和网状裂缝"""
    def __init__(self, threshold=3, HWratio=10, Histratio=0.5, alpha=15, beta=55):
        """
        初始化分类裂缝的参数
        :param threshold: 阈值，用于分类裂缝的阈值
        :param HWratio: 高宽比，用于分类裂缝的高宽比阈值
        :param Histratio: 直方图比例，用于分类裂缝的直方图比例阈值
        :param alpha: 倾斜裂纹分类的较低阈值角度
        :param beta: 倾斜裂纹分类的上阈值角度
        """
        self.threshold = threshold
        self.HWratio = HWratio
        self.Histratio = Histratio
        self.types = {0: 'Horizontal',
                      1: 'Vertical',
                      2: 'Oblique',
                      3: 'Mesh'}
        self.alpha = alpha
        self.beta = beta

    def inference_minAreaRect(self, mask):
        """
        从二值化掩膜图像中获取最小外接矩形的相关信息, 旋转矩形框长边与x轴的夹角.
        旋转角度 angle 是相对于图像水平方向的夹角，范围是 -90 到 +90 度.
        然而，一般情况下，我们习惯将角度定义为相对于 x 轴正方向的夹角，范围是 -180 到 +180 度.
        """
        contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        contour_merge = np.vstack(contours)
        minAreaRect = cv2.minAreaRect(contour_merge)
        w, h = minAreaRect[1]
        if w > h:
            angle = int(minAreaRect[2])
        else:
            angle = -(90 - int(minAreaRect[2]))

        return w, h, angle

    def hist_judge(self, hist_v, esp=1e-5):
        less_than_T = np.count_nonzero((hist_v > 0) & (hist_v <= self.threshold))
        more_than_T = np.count_nonzero(hist_v > self.threshold)
        return more_than_T / (less_than_T + esp) > self.Histratio

    def angle2cls(self, angle):
        angle = abs(angle)
        assert 0 <= angle <= 90, "ERROR: The angle value exceeds the limit and should be between 0 and 90 degrees!"
        if angle < self.alpha:
            return self.types[0]
        elif self.alpha <= angle <= self.beta:
            return self.types[2]
        elif angle > self.beta:
            return self.types[1]
        else:
            return None

    def skeleton_map(self, binary):
        """获取骨架图的信息, 因为只是用于判断类型, 所以不要求其是否有毛刺"""
        skimage = skeletonize(binary)
        skimage = morphology.closing(skimage, morphology.disk(3))
        skepoints = np.argwhere(skimage)
        skimage = skimage.astype(np.uint8) * 255
        return skimage, skepoints

    def classify(self, mask):
        """
        针对当前crack instance，对其进行分类；
        主要利用了骨骼点双向投影直方图、旋转矩形框宽高比 /角度；
        """
        H, W = mask.shape[:2]
        mask = binarization(mask)
        skeimage, skeleton_pts = self.skeleton_map(mask)
        w, h, angle = self.inference_minAreaRect(skeimage)
        if w / h < self.HWratio or h / w < self.HWratio:
            pts_y, pts_x = skeleton_pts[:, 0], skeleton_pts[:, 1]
            hist_x = np.histogram(pts_x, W)
            hist_y = np.histogram(pts_y, H)
            if self.hist_judge(hist_x[0]) and self.hist_judge(hist_y[0]):
                return self.types[3]

        return self.angle2cls(angle)


class detect_crack():
    def __init__(self, mask):
        self.mask_copy = mask.copy()
        self.crack_type = CrackType()
        self.overall_crack_category = self.crack_type.classify(mask)  # 判断整体类型
        self.total_length = 0
        self.max_wide = 0
        connected_thresh = self.preprocess_binary(mask)
        self.total_area = np.count_nonzero(connected_thresh)
        self.regions_info = self.get_label_region(connected_thresh)
        self.crack_summary = {
            "crack_category": self.overall_crack_category,  # 整体裂缝类别
            "total_area": self.total_area,  # 总面积
            "crack_number": len(self.regions_info),  # 裂缝个数
            "crack_length": self.total_length / 2,
            "crack_max_wide": self.max_wide
        }
        # 将每个区域的信息添加到字典中
        for i, region in enumerate(self.regions_info):
            region_key = f"Region{i+1}"
            self.crack_summary[region_key] = region

    def preprocess_binary(self, mask):
        # 二值化处理
        thresh = binarization(mask)
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
        # 将相邻点连接在一起，避免因为单个像素间断出现多条裂缝
        connected = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
        return connected

    def get_label_region(self, thresh, connectivity=2):
        # 获取标记的区域及其额外属性
        def region_incircle(regionmask, intensity_image=None):
            # 计算区域的内切圆直径
            max_val = 0
            max_dist_pt = None
            regionmask = bool2mask(regionmask)
            dist_transform = cv2.distanceTransform(regionmask, cv2.DIST_L2, 3)
            min_val, curr_max_val, _, curr_max_dist_pt = cv2.minMaxLoc(dist_transform)
            if curr_max_val > max_val:
                max_val = curr_max_val
                max_dist_pt = curr_max_dist_pt
            wide = max_val * 2
            return wide, max_dist_pt
        labeled_image = measure.label(thresh, connectivity=connectivity)
        region_props = measure.regionprops(labeled_image, extra_properties=(region_incircle,))
        regions_info = []
        for i, region in enumerate(region_props):
            region_info = {}
            region_info["crack_category"] = self.crack_type.angle2cls(90 - abs(np.degrees(region.orientation)))  # 根据方向计算裂缝类型
            region_info["area"] = region.area  # 记录区域的面积
            region_info["perimeter"] = region.perimeter  # 记录区域的周长
            region_info["orientation"] = 90 - abs(np.degrees(region.orientation))  # 记录区域的方向
            region_info["max_width"] = region.region_incircle[0]  # 记录内切圆的最大宽度
            minr, minc, maxr, maxc = region.bbox
            region_info["max_width_point"] = (
                region.region_incircle[1][0]+minc, region.region_incircle[1][1]+minr
            )  # 记录内切圆最大宽度的点
            region_info["bounding_box"] = [(minc, minr), (maxc, maxr)]  # 记录区域的边界框
            regions_info.append(region_info)
            self.total_length += region.perimeter
            if region.region_incircle[0] > self.max_wide:
                self.max_wide = region.region_incircle[0]

        return regions_info

    def summary(self):
        def format_summary(data, indent=0):
            """递归格式化字典为结构化的字符串"""
            formatted_str = ""
            for key, value in data.items():
                if isinstance(value, dict):
                    formatted_str += " " * indent + f"{key}:\n"
                    formatted_str += format_summary(value, indent + 4)
                else:
                    formatted_str += " " * indent + f"{key}: {value}\n"
            return formatted_str

        return format_summary(self.crack_summary)


    def plot(self, circlecolor=(0, 255, 0), bboxcolor=(255, 0, 0), textcolor=(0, 0, 255),
             font_scale=.7, thickness=1):
        """
        绘制带有内切圆、边界框和标注的区域图像
        :param circle_color: 内切圆的颜色 (默认绿色)
        :param bboxcolor: 边界框的颜色 (默认红色)
        :param textcolor: 标注文字的颜色 (默认黄色)
        :param font_scale: 标注文字的大小
        :param thickness: 文字与边界框的厚度
        :return: 绘制后的图像
        """
        plot_image = self.mask_copy.copy()
        for i, region in enumerate(self.regions_info):
            # 画内切圆
            max_width = region["max_width"]
            max_width_point = region["max_width_point"]
            cv2.circle(plot_image, max_width_point, int(max_width / 2), circlecolor, 1, 1, 0)
            # 画边界框, 标注区域类别
            bbox = region["bounding_box"]
            label = f"Region {i+1}"
            PutBoxText(plot_image, (bbox[0][0], bbox[0][1], bbox[1][0], bbox[1][1]), label, 0, bboxcolor, textcolor,
                       font_scale=font_scale, thickness=thickness, bboxthickness=thickness)
        return plot_image

if __name__=="__main__":
    img_mask = cv2.imread(r"E:\PythonProject\pyzjrPyPi\pyzjr\measure\fissure\0028.png")
    # crack_type = CrackType()
    # result = crack_type.classify(img_mask)
    decrack = detect_crack(img_mask)
    print(decrack.summary())
    plot_image = decrack.plot()
    # print(result)
    # cv2.imwrite("decrack.png", plot_image)
    cv2.imshow("Crack Detection", plot_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()



from .io import (
    imwrite, display, StackedImagesV1, StackedImagesV2, imattributes, imshowplt,
    StackedpltV1, StackedpltV2, matplotlib_patch, VideoCap, read_bgr, read_gray,
    read_rgb, read_tensor, read_image, imwrite, display, StackedImagesV1,
    StackedImagesV2, imattributes, image_to_bchw, bchw_to_image, tensor_to_image,
    image_to_tensor, imagelist_to_tensor
)
from .plot import (
    AddText, PutMultiLineText, PutMultiLineCenteredText, PutBoxText,
    PutRectangleText, DrawPolygon, DrawCornerRectangle, OverlayPng, ConvertBbox
)
from .colorspace import (
    to_gray, rgb2bgr, bgr2rgb, to_hsv, hsv2rgb, hsv2bgr, pil2cv, cv2pil, create_palette
)
from .core import Timer, FPS, Runcodes, timing
from .printf import (
    ConsoleLogger, redirect_console, colorstr, colorfulstr, show_config, LoadingBar,
    printprocess, printlog, printcolor, printshape
)


"""
Copyright (c) 2024, Auorui.
All rights reserved.
Time: 2024-01-12 16:14

组合池化, 并且有  ’ 可选择类型 ‘
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

def adaptive_pool2d(input_tensor, output_size, pool_type='max'):
    """
    两种选择型自适应池化，'max' 和 'avg'

    Args:
        - input_tensor: 输入张量
        - output_size: 输出尺寸
        - pool_type: 池化类型，可以是 'max' 或 'avg'

    Returns:
        - output_tensor: 池化后的张量
    """
    if pool_type == 'max':
        pool_func = F.adaptive_max_pool2d
    elif pool_type == 'avg':
        pool_func = F.adaptive_avg_pool2d
    else:
        raise ValueError("Unsupported pooling type. Use 'max' or 'avg'.")

    output_tensor = pool_func(input_tensor, output_size)
    return output_tensor

def adaptive_avgmax_pool2d(x, output_size):
    """
    两种选择型自适应池化，'max' 和 'avg'的平均值

    Args:
        - x: 输入张量
        - output_size: 输出尺寸

    Returns:
        - 池化后的张量的平均值
    """
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return 0.5 * (x_avg + x_max)

def adaptive_catavgmax_pool2d(x, output_size):
    """
    两种选择型自适应池化，'max' 和 'avg'的cat连接

    Args:
        - x: 输入张量
        - output_size: 输出尺寸

    Returns:
        - 连接池化后的张量
    """
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return torch.cat((x_avg, x_max), dim=1)

class AdaptiveAvgMaxPool2d(nn.Module):
    """
    自适应平均最大池化的PyTorch模块

    Args:
        - output_size: 输出尺寸
    """
    def __init__(self, output_size=1):
        super(AdaptiveAvgMaxPool2d, self).__init__()
        self.output_size = output_size

    def forward(self, x):
        return adaptive_avgmax_pool2d(x, self.output_size)

class AdaptiveCatAvgMaxPool2d(nn.Module):
    """
    自适应连接平均最大池化的PyTorch模块

    Args:
        - output_size: 输出尺寸
    """
    def __init__(self, output_size=1):
        super(AdaptiveCatAvgMaxPool2d, self).__init__()
        self.output_size = output_size

    def forward(self, x):
        return adaptive_catavgmax_pool2d(x, self.output_size)

class FastAdaptiveAvgPool2d(nn.Module):
    """
    自定义的自适应平均池化层

    Args:
        - flatten: 是否对结果进行扁平化
    """
    def __init__(self, flatten=False):
        super(FastAdaptiveAvgPool2d, self).__init__()
        self.flatten = flatten

    def forward(self, x):
        return x.mean((2, 3)) if self.flatten else x.mean((2, 3), keepdim=True)


class SelectAdaptivePool2d(nn.Module):
    """
    Selectable global pooling layer with dynamic input kernel size
    """
    def __init__(self, output_size=1, pool_type='fast', flatten=False):
        super(SelectAdaptivePool2d, self).__init__()
        self.pool_type = pool_type or ''  # convert other falsy values to empty string for consistent TS typing
        self.flatten = flatten
        if pool_type == '':
            self.pool = nn.Identity()  # pass through
        elif pool_type == 'fast':
            assert output_size == 1
            self.pool = FastAdaptiveAvgPool2d(self.flatten)
        elif pool_type == 'avg':
            self.pool = nn.AdaptiveAvgPool2d(output_size)
        elif pool_type == 'max':
            self.pool = nn.AdaptiveMaxPool2d(output_size)
        elif pool_type == 'avgmax':
            self.pool = AdaptiveAvgMaxPool2d(output_size)
        elif pool_type == 'catavgmax':
            self.pool = AdaptiveCatAvgMaxPool2d(output_size)
        else:
            assert False, 'Invalid pool type: %s' % pool_type

    def forward(self, x):
        x = self.pool(x)
        return x


if __name__ == "__main__":
    input_tensor = torch.rand((1, 3, 32, 32))
    output_max = adaptive_pool2d(input_tensor, (1, 1), pool_type='max')
    output_avg = adaptive_pool2d(input_tensor, (1, 1), pool_type='avg')
    output_avgmax = adaptive_avgmax_pool2d(input_tensor, (1, 1))
    output_catavgmax = adaptive_catavgmax_pool2d(input_tensor, (1, 1))
    output_fast = SelectAdaptivePool2d()(input_tensor)
    print("自适应最大池化结果：\n", output_max.shape)
    print("\n自适应平均池化结果：\n", output_avg.shape)
    print("\n自适应平均池化与自适应最大池化的结合结果：\n", output_avgmax.shape)
    print("\n自适应平均池化与自适应最大池化的连接结果：\n", output_catavgmax.shape)
    print("\n自定义的自适应平均池化层:\n", output_fast.shape)


import os
import shutil
from pyzjr.data.utils.getpath import getPhotoPath

def batch_rename_images(
        target_path,
        save_path,
        start_index=None,
        newprefix=None,
        newsuffix=None,
        num_type=1,
        message=True
):
    """
    重命名图像文件夹中的所有图像文件并保存到指定文件夹
    :param target_path: 目标文件路径
    :param save_path: 文件夹的保存路径
    :param start_index: 默认为 1, 从多少号开始
    :param newprefix: 重命名的通用格式前缀, 如 rename001.png, rename002.png...
    :param newsuffix: 重命名的通用格式后缀, 如 001rename.png, 002rename.png...
    :param num_type: 数字长度, 比如 3 表示 005
    :param message: 是否打印修改信息
    """
    os.makedirs(save_path, exist_ok=True)
    images_paths, _ = getPhotoPath(target_path)
    current_num = start_index if start_index is not None else 1

    for i, image_path in enumerate(images_paths):
        image_name = os.path.basename(image_path)
        name, ext = os.path.splitext(image_name)

        padded_i = str(current_num).zfill(num_type)
        if newprefix and newsuffix:
            new_image_name = f"{newprefix}{padded_i}{newsuffix}{ext}"
        elif newprefix:
            new_image_name = f"{newprefix}{padded_i}{ext}"
        elif newsuffix:
            new_image_name = f"{padded_i}{newsuffix}{ext}"
        else:
            new_image_name = f"{padded_i}{ext}"

        new_path = os.path.join(save_path, new_image_name)
        current_num += 1
        if message:
            print(f"{i + 1} Successfully rename {image_path} to {new_path}")
        shutil.copy(image_path, new_path)

    print("Batch renaming and saving of files completed!")

def convert_suffix(file_path, format):
    """
    将文件路径中的后缀名转换为新的后缀名

    Args:
        file_path (str): 要转换后缀名的文件路径。
        format (str): 新的后缀名，不需要包含点（.）。

    Returns:
        str: 转换后的文件路径。
    """
    base_name, ext = os.path.splitext(file_path)
    new_file_path = base_name + '.' + format
    return new_file_path

def modify_images_suffix(target_path, format='png', message=True):
    """批量修改图像文件后缀"""
    images_paths, _ = getPhotoPath(target_path)
    for i, one_image_path in enumerate(images_paths):
        new_path = convert_suffix(one_image_path, format)
        os.rename(one_image_path, new_path)
        if message:
            print(f"Converting {one_image_path} to {new_path}")

if __name__=="__main__":
    target_path = './GT'
    save_path = './GT1'

    batch_rename_images(target_path, save_path, start_index=35, newprefix='root', newsuffix='ok', num_type=1)
    modify_images_suffix(save_path, format='jpg')


import os
import shutil
import math
import numpy as np
from PIL import Image
from pyzjr.data.utils.tuplefun import to_2tuple
from pyzjr.data.utils.getpath import getPhotoPath

def modify_images_size(target_path, target_hwsize, save_path=None, message=True):
    """批量修改图像大小"""
    h, w = to_2tuple(target_hwsize)
    images_paths, _ = getPhotoPath(target_path)
    os.makedirs(save_path, exist_ok=True)
    for i, one_image_path in enumerate(images_paths):
        try:
            image = Image.open(one_image_path)
            resized_image = image.resize((w, h))
            base_name = os.path.basename(one_image_path)
            if save_path is not None:
                new_path = os.path.join(save_path, base_name)
            else:
                new_path = one_image_path

            resized_image.save(new_path)
            if message:
                print(f"Resized {one_image_path} to {new_path}")
        except Exception as e:
            print(f"Error resizing {one_image_path}: {e}")

def split_train_val_txt(target_path, train_ratio=.8, val_ratio=.2, onlybasename=False):
    """
    如果 train_ratio + val_ratio = 1 表示只划分训练集和验证集, train_ratio + val_ratio < 1
    表示将剩余的比例划分为测试集
    """
    assert train_ratio + val_ratio <= 1
    test_ratio = 1. - (train_ratio + val_ratio)
    images_paths, _ = getPhotoPath(target_path)

    num_images = len(images_paths)
    num_train = round(num_images * train_ratio)
    num_val = num_images - num_train if test_ratio == 0 else math.ceil(num_images * val_ratio)
    num_test = 0 if test_ratio == 0 else num_images - (num_train + num_val)

    with open(os.path.join(target_path, 'train.txt'), 'w') as train_file, \
            open(os.path.join(target_path, 'val.txt'), 'w') as val_file, \
            open(os.path.join(target_path, 'test.txt'), 'w') as test_file:
        for i, image_path in enumerate(images_paths):
            if onlybasename:
                image_name, _ = os.path.splitext(os.path.basename(image_path))
            else:
                image_name = image_path
            if i < num_train:
                train_file.write(f"{image_name}\n")
            elif i < num_train + num_val:
                val_file.write(f"{image_name}\n")
            else:
                test_file.write(f"{image_name}\n")

    print(f"Successfully split {num_images} images into {num_train} train, {num_val} val, and {num_test} test.")

def copy_images_to_directory(target_path, save_folder, message=True):
    """复制整个文件夹（图像）到另外一个文件夹"""
    try:
        os.makedirs(save_folder, exist_ok=True)
        source_path, _ = getPhotoPath(target_path)
        for img_path in source_path:
            base_file_name = os.path.basename(img_path)
            destination_path = os.path.join(save_folder, base_file_name)
            shutil.copy2(img_path, destination_path)
            if message:
                print(f"Successfully copied folder: {img_path} to {save_folder}")
    except Exception as e:
        print(f"Error copying folder, {e}")

def get_images_mean_std(target_path):
    """返回RGB顺序的mean与std"""
    images_paths, _ = getPhotoPath(target_path)
    num_images = len(images_paths)
    mean_sum = np.zeros(3)
    std_sum = np.zeros(3)

    for one_image_path in images_paths:
        pil_image = Image.open(one_image_path).convert("RGB")
        img_asarray = np.asarray(pil_image) / 255.0
        individual_mean = np.mean(img_asarray, axis=(0, 1))
        individual_stdev = np.std(img_asarray, axis=(0, 1))
        mean_sum += individual_mean
        std_sum += individual_stdev

    mean = mean_sum / num_images
    std = std_sum / num_images
    return mean.astype(np.float32), std.astype(np.float32)



if __name__=="__main__":
    target_path = ...
    save_path = ...
    # modify_images_suffix(target_path, 'png')
    # modify_images_size(target_path, (420, 512), './test2')
    mean, std = get_images_mean_std(target_path)
    print(mean, std)
    # copy_images_to_directory(target_path, save_path)
    # split_train_val_txt(target_path, val_ratio=.2,onlybasename=True)
    # batch_rename_images(target_path, save_path, start_number=None)


import torch
import numpy as np
from pyzjr.utils.check import is_numpy

def image_to_bchw(image_data):
    """
    将图像加载为可输入网络的形状 b c h w, 且 b = 1
    """
    image_np = np.array(image_data).astype(np.uint8) / 255
    image_bchw = np.expand_dims(np.transpose(image_np[:, :, ::-1], (2, 0, 1)), 0)
    return image_bchw

def bchw_to_image(image_bchw):
    """
    将网络输出转为图像类型, 且 b = 1
    """
    image = image_bchw.detach().cpu().squeeze(0)
    image = image.permute(1, 2, 0).numpy()
    image = (image - image.min()) / (image.max() - image.min())
    image = (image[:, :, ::-1] * 255).astype(np.uint8)
    return image

def image_to_tensor(image, keepdim=True):
    """
    Convert numpy images to PyTorch 4d tensor images
    'keepdim' indicates whether to maintain the current dimension, otherwise it will be changed to type 4d
    Example:
        img = np.ones((3, 3))
        image_to_tensor(img).shape
    [1, 3, 3]
        img = np.ones((4, 4, 1))
        image_to_tensor(img).shape
    [1, 4, 4]
        img = np.ones((4, 4, 3))
        image_to_tensor(img, keepdim=False).shape
    [1, 3, 4, 4]
    """
    if is_numpy(image):
        if len(image.shape) > 4 or len(image.shape) < 2:
            raise ValueError("Input size must be a two, three or four dimensional array")
    input_shape = image.shape
    tensor = torch.from_numpy(image)

    if len(input_shape) == 2:
        # (H, W) -> (1, H, W)
        tensor = tensor.unsqueeze(0)
    elif len(input_shape) == 3:
        # (H, W, C) -> (C, H, W)
        tensor = tensor.permute(2, 0, 1)
    elif len(input_shape) == 4:
        # (B, H, W, C) -> (B, C, H, W)
        tensor = tensor.permute(0, 3, 1, 2)
        keepdim = True  # no need to unsqueeze
    else:
        raise ValueError(f"Cannot process image with shape {input_shape}")

    return tensor if keepdim else tensor.unsqueeze(0)

def tensor_to_image(tensor, keepdim = False):
    """Convert PyTorch tensor image to numpy image
    Returns:
        image of the form :math:`(H, W)`, :math:`(H, W, C)` or :math:`(B, H, W, C)`.
    Example:
        img = torch.ones(1, 3, 3)
        tensor_to_image(img).shape
    (3, 3)
        img = torch.ones(3, 4, 4)
        tensor_to_image(img).shape
    (4, 4, 3)
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Input type is not a Tensor. Got {type(tensor)}")

    if len(tensor.shape) > 4 or len(tensor.shape) < 2:
        raise ValueError("Input size must be a two, three or four dimensional tensor")

    input_shape = tensor.shape
    image = tensor.cpu().detach().numpy()

    if len(input_shape) == 2:
        # (H, W) -> (H, W)
        pass
    elif len(input_shape) == 3:
        # (C, H, W) -> (H, W, C)
        if input_shape[0] == 1:
            # Grayscale for proper plt.imshow needs to be (H,W)
            image = image.squeeze()
        else:
            image = image.transpose(1, 2, 0)
    elif len(input_shape) == 4:
        # (B, C, H, W) -> (B, H, W, C)
        image = image.transpose(0, 2, 3, 1)
        if input_shape[0] == 1 and not keepdim:
            image = image.squeeze(0)
        if input_shape[1] == 1:
            image = image.squeeze(-1)
    else:
        raise ValueError(f"Cannot process tensor with shape {input_shape}")

    return image

def imagelist_to_tensor(imagelist):
    """Converts a list of numpy images to a PyTorch 4d tensor image.
    Args:
        images: list of images, each of the form :math:`(H, W, C)`.
        Image shapes must be consistent
    Returns:
        tensor of the form :math:`(B, C, H, W)`.
    Example:
        imgs = [np.ones((4, 4, 1)),
                np.zeros((4, 4, 1))]
        image_list_to_tensor(imgs).shape
    torch.Size([2, 1, 4, 4])
    """
    if len(imagelist[0].shape) != 3:
        raise ValueError("Input images must be three dimensional arrays")
    list_of_tensors = []
    for image in imagelist:
        list_of_tensors.append(image_to_tensor(image))
    return torch.stack(list_of_tensors)




from .classification import (
    ClassificationIndex,
    ConfusionMatrixs,
    BinaryConfusionMatrix,
    MulticlassConfusionMatrix,
    calculate_metrics
)
from .segmentation import (
    calculate_confusion_matrix_multilabel,
    SegmentationIndex,
)
from .infer_testset import (
    calculate_area,
    mean_iou,
    auc_roc,
    accuracy,
    dice,
    kappa,
    InferTestset
)
from .utils import (
    generate_class_weights
)


from .image_script import (
    modify_images_size,
    copy_images_to_directory,
    split_train_val_txt,
    get_images_mean_std,
)

from .rename import (
    batch_rename_images,
    convert_suffix,
    modify_images_suffix
)


"""
去雾数据集专用格式如下:
    - rshazy
        - test
        - train
            - GT
                1.png
                2.png
            - hazy
                1.png
                2.png
                ...
        - test.txt
            1
            2
            ...
        - train.txt
"""
import os
import cv2
import random
import numpy as np
from pyzjr.data.utils.tuplefun import to_2tuple
from pyzjr.data.datasets.baseset import BaseDataset


class DeHazeDataset(BaseDataset):
    def __init__(
            self,
            root_dir,  # Synscapes
            target_shape,
            is_train=True,
            edge_decay=0,
            only_h_flip=False
    ):
        super(DeHazeDataset, self).__init__()
        self.mode = is_train
        self.edge_decay = edge_decay
        self.only_h_flip = only_h_flip
        self.target_shape = to_2tuple(target_shape)
        self.data_dir = os.path.join(root_dir, 'train' if is_train else 'test')
        txt_path = self.data_dir + '.txt'
        self.txt_content = self.read_txt(txt_path)

    def __len__(self):
        return len(self.txt_content)

    def __getitem__(self, item):
        cv2.setNumThreads(0)
        cv2.ocl.setUseOpenCL(False)

        img_name = f"{self.txt_content[item]}.png"
        # 归一化到 [-1, 1]
        source_img = self.read_image(os.path.join(self.data_dir, 'hazy', img_name)) * 2 - 1
        target_img = self.read_image(os.path.join(self.data_dir, 'GT', img_name)) * 2 - 1

        if self.mode:
            [source_img, target_img] = self.augment([source_img, target_img], self.target_shape, self.edge_decay, self.only_h_flip)
        else:
            [source_img, target_img] = self.align([source_img, target_img], self.target_shape)

        return self.hwc2chw(source_img), self.hwc2chw(target_img)


    def augment(self, imglists, size, edge_decay=0., only_h_flip=False):
        H, W, _ = imglists[0].shape
        Hc, Wc = size

        # simple re-weight for the edge
        if random.random() < Hc / H * edge_decay:
            Hs = 0 if random.randint(0, 1) == 0 else H - Hc
        else:
            Hs = random.randint(0, H - Hc)

        if random.random() < Wc / W * edge_decay:
            Ws = 0 if random.randint(0, 1) == 0 else W - Wc
        else:
            Ws = random.randint(0, W - Wc)

        for i in range(len(imglists)):
            imglists[i] = imglists[i][Hs:(Hs + Hc), Ws:(Ws + Wc), :]

        # horizontal flip
        if random.randint(0, 1) == 1:
            for i in range(len(imglists)):
                imglists[i] = np.flip(imglists[i], axis=1).copy()

        if not only_h_flip:
            # bad data augmentations for outdoor
            rot_deg = random.randint(0, 3)
            for i in range(len(imglists)):
                imglists[i] = np.rot90(imglists[i], rot_deg, (0, 1)).copy()

        return imglists


    def align(self, imglists, size):
        H, W, _ = imglists[0].shape
        Hc, Wc = size

        Hs = (H - Hc) // 2
        Ws = (W - Wc) // 2
        for i in range(len(imglists)):
            imglists[i] = imglists[i][Hs:(Hs + Hc), Ws:(Ws + Wc), :]

        return imglists


if __name__=="__main__":
    from torch.utils.data import DataLoader
    data_dir = r"rshazy"
    train_dataset = DeHazeDataset(data_dir, target_shape=256, is_train=True)
    val_dataset = DeHazeDataset(data_dir, target_shape=256,
                                is_train=False)
    train_loader = DataLoader(train_dataset,
                              batch_size=1,
                              shuffle=True,
                              num_workers=2,
                              pin_memory=True,
                              drop_last=True)
    val_loader = DataLoader(val_dataset,
                            batch_size=1,
                            num_workers=2,
                            pin_memory=True)
    for i , (gt, hazy) in enumerate(train_loader):
        print(i, gt.shape, hazy.shape)
    for i , (gt, hazy) in enumerate(val_loader):
        print(i, gt.shape, hazy.shape)


import torch
import numpy as np

IMAGENET_MEAN = 0.485, 0.456, 0.406  # RGB mean
IMAGENET_STD = 0.229, 0.224, 0.225  # RGB standard deviation

def imagenet_denormal(image):
    # 对RGB图像x进行ImageNet反规范化, i.e. = x * std + mean
    return image * IMAGENET_STD + IMAGENET_MEAN

def imagenet_normal(image):
    # 对RGB图像x进行ImageNet规范化, i.e. = (x - mean) / std
    image = image.astype('float') / 255
    normal_image = (image - IMAGENET_MEAN) / IMAGENET_STD
    return normal_image

def min_max_normal(np_image):
    """
    Min-Max normalization, scales pixel values to a specific range [0, 1].
    """
    min_val = np.min(np_image)
    max_val = np.max(np_image)
    min_and_max = (np_image - min_val) / (max_val - min_val)
    return min_and_max

def z_score_normal(np_image):
    """
    Z-score, computes the standard scores of given data.
    """
    mean = np.mean(np_image)
    std = np.std(np_image)
    z_scores = (np_image - mean) / std
    return z_scores

def linear_normal(np_image):
    """
    Linear normalization
    http://en.wikipedia.org/wiki/Normalization_%28image_processing%29
    """
    arr = np_image.astype('float')
    for i in range(3):
        minval = arr[..., i].min()
        maxval = arr[..., i].max()
        if minval != maxval:
            arr[..., i] -= minval
            arr[..., i] *= (255.0 / (maxval - minval))
    return arr

def zero_centered_normal(np_image):
    """Zero-centered (centered)"""
    np_image = np_image.astype('float32')
    mean = np.mean(np_image)
    np_image -= mean
    return np_image

class Normalizer():
    def __init__(self, normal_type='imagenet'):
        self.normal_type = normal_type

    def __call__(self, image):
        if isinstance(image, np.ndarray):
            self.image = image
        else:
            self.image = np.array(image)

        if self.normal_type == 'imagenet':
            return imagenet_normal(self.image)
        elif self.normal_type == 'min_max':
            return min_max_normal(self.image)
        elif self.normal_type == 'z_score':
            return z_score_normal(self.image)
        elif self.normal_type == 'linear':
            return linear_normal(self.image)
        elif self.normal_type == 'zero_centered':
            return zero_centered_normal(self.image)
        else:
            raise ValueError(f"Unsupported normalization type: {self.normal_type}, "
                             f"You can choose to use these parameters"
                             f"' imagenet ',' min_max ',' z_score ',' linear ', and ' zero_centered '")


def denormalize(image, mean, std):
    """
    Denormalization operation, converts normalized data back to the original range.

    Parameters:
        - image: Input data (NumPy array or Torch tensor)
        - mean: Mean used for normalization
        - std: Standard deviation used for normalization

    Returns:
        - Denormalized data
    """
    if isinstance(image, np.ndarray):
        # NumPy arrays
        denormalized_data = image * std + mean
    elif torch.is_tensor(image):
        # Torch tensors
        denormalized_data = image.clone()
        for d, m, s in zip(denormalized_data, mean, std):
            d.mul_(s).add_(m)
    else:
        raise ValueError("Unsupported input type. Only NumPy arrays and Torch tensors are supported.")

    return denormalized_data


if __name__ == "__main__":
    import torchvision.transforms as transforms
    from PIL import Image

    image_path = r"../../data/scripts/fire.jpg"
    image = Image.open(image_path)

    to_tensor = transforms.ToTensor()
    torch_image = to_tensor(image)

    imagenet_normalized = Normalizer('imagenet')
    min_max_normalized = Normalizer('min_max')
    z_score_normalized = Normalizer('z_score')
    linear_normalized = Normalizer('linear')
    zero_centered = Normalizer('zero_centered')


    print(imagenet_normalized(image))
    print(min_max_normalized(image))
    print(z_score_normalized(image))
    print(linear_normalized(image))
    print(zero_centered(image))


import os
import torch
from collections import OrderedDict

_torch_save = torch.save  # copy to avoid recursion errors
best_val_loss = float('inf')
init_metrics = float('-inf')

def _process_net_state_dict(net, use_devices=True):
    if torch.cuda.device_count() > 1 and use_devices:
        return net.module.state_dict()
    else:
        return net.state_dict()

def SaveModelPthSimplify(net, save_dir, epoch, save_period=30, use_devices=True):
    """
    Save the model according to training rounds

    Args:
        net: model to be saved
        save_dir: path that the model would be saved
        epoch: current epoch
        save_period: The frequency of saving the model during training.
                    The default is 10.
    """
    net = _process_net_state_dict(net, use_devices)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    if epoch % save_period == 0:
        try:
            _torch_save(net, os.path.join(save_dir, f"model_epoch_{epoch}.pth"))
            print(f'\033[34mSave model to {os.path.join(save_dir, f"model_epoch_{epoch}.pth")}')
        except Exception as e:
            print(f"\033[31mFailed to save model at epoch {epoch}: {e}")


def SaveModelPthBestloss(net, save_dir, val_loss, epoch, save_period=None, use_devices=True):
    """
    Save the model based on training rounds (optional),
    and save the best model based on validation loss.

    Args:
        net: model to be saved
        save_dir: path that the model would be saved
        val_loss (float): Verification losses for the current period. Usually, certain
                        indicators can also be used, such as dice, accuracy, etc.
        epoch: current epoch
        save_period (int, optional): The frequency of saving the model during training.
                                    The default is None.
    """
    global best_val_loss
    os.makedirs(save_dir, exist_ok=True)
    net = _process_net_state_dict(net, use_devices)
    if epoch % save_period == 0 and save_period is not None:
        _torch_save(net,
                   os.path.join(save_dir, f'epoch{epoch}_loss{val_loss:.4}.pth'))
        print(f"\033[34mmodel saved at epoch —— {epoch}")
    if val_loss < best_val_loss:
        best_val_loss = val_loss
        best_model_path = os.path.join(save_dir, "best_model.pth")
        _torch_save(net, best_model_path)
        print(f"\033[34mBest model saved at epoch —— {epoch} loss —— {val_loss}")
        print(f'\033[34mSave best model to {best_model_path}')


def SaveModelPthBestMetrics(net, save_dir, metric, epoch, minlimit=None, use_devices=True):
    """
    Save the model if the current metric is the best observed so far.

    Args:
        net: model to be saved.
        save_dir: path that the model would be saved.
        metric (float): The current metric value.
        epoch (int): The current epoch.
        minlimit (float, optional): The minimum threshold for the metric. If the
                                     metric is less than or equal to `minlimit`,
                                     the model will not be saved.
    """
    global init_metrics
    os.makedirs(save_dir, exist_ok=True)
    net = _process_net_state_dict(net, use_devices)
    if minlimit is not None and metric <= minlimit:
        print(f"\033[33mMetric {metric} is below the minimum limit of {minlimit}, model will not be saved this epoch.")
        return
    if epoch <= 1 or metric > init_metrics:
        init_metrics = metric
        model_path = os.path.join(save_dir, f'best_model.pth')
        _torch_save(net, model_path)
        print(f"\033[34mBest model saved at epoch —— {epoch}, metric —— {init_metrics}")
        print(f'\033[34mSave best model to {model_path}')


def SaveModelPth(net, save_dir, epoch, total_epochs=None, val_loss=None, metric=None, save_period=80, use_devices=True):
    """
    Save the model at specified intervals or if the model achieves the best validation loss/metric.
    Also saves the model at the last epoch.

    Args:
        net: model to be saved
        save_dir: path that the model would be saved
        epoch: current epoch
        total_epochs (int, optional): Total number of epochs. If provided, the model will be saved at the last epoch.
        val_loss (float, optional): Validation loss for the current epoch.
        metric (float, optional): The current metric value.
        save_period (int, optional): The frequency of saving the model during training.
    """
    global best_val_loss, init_metrics
    os.makedirs(save_dir, exist_ok=True)
    net = _process_net_state_dict(net, use_devices)
    # Save model at specified intervals
    if epoch % save_period == 0:
        try:
            torch.save(net, os.path.join(save_dir, f"model_epoch_{epoch}.pth"))
            print(f'\033[34mModel saved at epoch {epoch} to {os.path.join(save_dir, f"model_epoch_{epoch}.pth")}')
        except Exception as e:
            print(f"\033[31mFailed to save model at epoch {epoch}: {e}")

    # Save model if it achieves the best validation loss
    if val_loss is not None and val_loss < best_val_loss:
        best_val_loss = val_loss
        best_loss_model_path = os.path.join(save_dir, "best_loss_model.pth")
        torch.save(net, best_loss_model_path)
        print(f"\033[34mBest model saved at epoch —— {epoch} loss —— {val_loss}")
        print(f'\033[34mSave best model to {best_loss_model_path}')

    # Save model if it achieves the best metric
    if metric is not None:
        if metric > init_metrics:
            init_metrics = metric
            best_metric_model_path = os.path.join(save_dir, "best_metric_model.pth")
            torch.save(net, best_metric_model_path)
            print(f"\033[34mBest model saved at epoch —— {epoch}, metric —— {init_metrics}")
            print(f'\033[34mBest model saved to {best_metric_model_path}')

    # Save model at the last epoch
    if total_epochs is not None and epoch == total_epochs:
        last_model_path = os.path.join(save_dir, "last.pth")
        torch.save(net, last_model_path)
        print(f"\033[34mModel saved at the last epoch ({epoch}) to {last_model_path}")


def load_partial_weights(net, pretrained_path=None, pretrained_dict=None, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu'), use_devices=True):
    """
    加载部分权重到模型中，只加载匹配的权重，并且返回更新后的模型。

    Args:
        net (torch.nn.Module): 要加载部分权重的当前模型。
        pretrained_path (str, optional): 预训练权重文件的路径。如果提供了此路径，将从该路径加载权重。默认为None。
        pretrained_dict (dict, optional): 预训练的权重字典。如果提供了此字典，则直接从该字典加载权重，忽略pretrained_path。默认为None。

    Returns:
        torch.nn.Module: 加载了部分权重的模型。
    """
    net = net.to(device)
    model_dict = _process_net_state_dict(net, use_devices)
    if pretrained_path is not None:
        pretrained_dict = torch.load(pretrained_path, map_location=device)
    else:
        pretrained_dict = pretrained_dict
    # 创建存储匹配、未匹配和缺失权重的列表
    load_key, no_load_key, missing_keys, temp_dict = [], [], [], {}

    for k, v in pretrained_dict.items():
        if k in model_dict:
            if model_dict[k].shape == v.shape:
                temp_dict[k] = v
                load_key.append(k)
            else:
                no_load_key.append(k)
                print(f"\033[31mKey: {k}, Model Shape: {model_dict[k].shape}, Pretrained Shape: {v.shape}")
        else:
            missing_keys.append(k)
            print(f"\033[31mKey: {k} not in model_dict.")

    model_dict.update(temp_dict)
    net.load_state_dict(model_dict)
    if not no_load_key and not missing_keys:
        print(f"\033[34m\nAll {len(load_key)} keys were successfully loaded.")
    else:
        print(f"\033[31m\nSuccessful Load Keys:{str(load_key)[:300]}……\n"
              f"Successful Load Key Num:{len(load_key)}")
        print(f"\033[31m\nFail To Load Keys:{str(no_load_key)[:100]}……\n"
              f"Fail To Load Key Num:{len(no_load_key)}")
        print(f"\033[31m\nMissing Keys:{str(missing_keys)[:100]}……\n"
              f"Missing Key Num:{len(missing_keys)}")
    return net


def save_checkpoint(net=None, optimizer=None, epoch=None, train_losses=None, train_metric=None, val_loss=None,
                    val_metric=None, check_loss=None, savepath=None, model_name=None, use_devices=True):
    """
    保存模型检查点，包括模型权重、优化器状态、训练损失、验证损失等信息。
    """
    net_weights = _process_net_state_dict(net, use_devices)
    save_json = {
        'net_state_dict': net_weights,
        'optimizer_state_dict': optimizer.state_dict(),
        'epoch': epoch,
        'train_losses': train_losses,
        'train_metric': train_metric,
        'val_loss': val_loss,
        'val_metric': val_metric
    }
    if  val_loss < check_loss:
        savepath = savepath + '/{}_best_params.pkl'.format(model_name)
        check_loss = val_loss
    else:
        savepath = savepath + '/{}_epoch_{}.pkl'.format(model_name, epoch)
    torch.save(save_json, savepath)
    print("checkpoint of {}th epoch saved at {}".format(epoch, savepath))

    return check_loss


def load_checkpoint(model=None, optimizer=None, checkpoint_path=None,  losses_flag=False):
    """
    从检查点文件中加载模型权重、优化器状态和训练信息。
    """
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint['net_state_dict'])
    if optimizer:
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    start_epoch = checkpoint['epoch']
    if losses_flag:
        losses = checkpoint['train_losses']
        return model, optimizer, start_epoch, losses
    else:
        return model, optimizer, start_epoch

def load_single_weights(model, model_path):
    state_dict = torch.load(model_path)
    new_state_dict = OrderedDict()

    for k, v in state_dict.items():
        name = k[7:]
        new_state_dict[name] = v
    model.load_state_dict(new_state_dict)
    return model


from .check import (
    is_tensor, is_pil, is_numpy, is_gray_image, is_rgb_image, is_parallel, is_Iterable, is_str,
    is_int, is_float, is_bool, is_list, is_tuple, is_list_or_tuple, is_none, is_not_none,
    is_positive_int, is_nonnegative_int, is_ascii, is_url, is_image_extension, is_video_extension,
    is_file,is_directory, is_directory_not_empty, is_path_exists, is_windows, is_linux,
    is_odd, is_even, get_image_size, get_image_num_channels
)
from .mathfun import (
    EuclidDistance, ChessboardDistance, CityblockDistance, CenterPointDistance, normal,
    gaussian2d, retain, round_up, round_down, factorial, pow, sqrt, rsqrt, cos, sin, tan,
    cot, sec, csc, arccos, arcsin, arctan, angle_to_2pi, to_degree, to_radians, exp, log_e,
    sinh, arsinh, cosh, arcosh, tanh, artanh, sigmoid, relu
)
from .torch_np import (
    to_numpy, to_tensor, hwc2chw, chw2hwc, to_bchw, allclose, moveaxis, in1d, clip, where,
    argwhere, argsort, nonzero, floor_divide, unravel_index, unravel_indices, ravel, any_np_pt,
    maximum, concatenate, cumsum, isfinite, searchsorted, repeat, isnan, ascontiguousarray,
    stack, unique, max, min, median, mean, std,
)
from .randfun import (
    rand, randint, randbool, randfluctuate, randlist, randintlist, randnormal, randshuffle,
    randstep, randstring, rand_choice_weighted
)